// PHYSICAL CALCULUS
// 
// Simple implementation of systems of physical units and of calculus based on physical quantities.
// 
// 11.16.09 - Denis MARION (den@melix.net)
///////////////////////////////////////////////////////////
funcprot(0);

////////
function x = defUnitSystem(NewSys, units, unit_names)
// defines a reference or a secondary system of units in the global scope.
if ~exists("SystemList")
  error("You must first define a global variable named ''SystemList'' in the scope.");
end;
global SystemList;
if (length(units) ~= size(unit_names',1))
  error("arguments #3 and #4 must be vectors of identical size.");
end;
if (isempty(SystemList))
  warning(sprintf("Reference system of units ''%s'' created. All further systems must declare %i different units.", NewSys, length(units)));
  x = mlist(["UNITSYS", "name", "conversionFactors", "unitNames"], NewSys, units, unit_names);
  SystemList($ + 1) = NewSys ;
  return;
else
  StdSys = UNITSYS(SystemList(1));
  if (length(StdSys.conversionFactors) ~= length(units))
    error(sprintf("All secondary systems must declare the same number of units as the reference system ...
      of units (here, %i)", length(units))); 
  end;  
end;
found = %F;
  for k = 1 : length(SystemList)
  if (SystemList(k) == NewSys)
    found = %T;
    break;
  end;
end;
if found
  errMsg = "Error: System of units is already defined in scope, clear it first to redefine it.";
  error(errMsg);
end ;
x = mlist(["UNITSYS", "name", "conversionFactors", "unitNames"], NewSys, units, unit_names);
SystemList($ + 1) = NewSys ;
endfunction ;


//////////////////////////
// OPERATOR OVERLOADING //
//////////////////////////
// MULTIPLICATION
///////
function x = %UNIT_m_s(a, b)
// implements the multiplication between a unit-vector and a const-vector
if ((length(a.values) == 1) | (length(b) == 1)) 
  x = mlist(["UNIT", "unitSystem", "values", "dimension"], a.unitSystem, a.values * b , a.dimension);
elseif (length(a.values) == length(b)) 
  x = mlist(["UNIT", "unitSystem", "values", "dimension"], a.unitSystem, a.values * b, a.dimension);
else
  error("Factors must have the same length (or one must be a singleton).");
end;
endfunction ;

//////
deff("x = %UNIT_s(a)","x = mlist([""UNIT"", ""unitSystem"", ""values"", ""dimension""], a.unitSystem, -a.values, a.dimension);"); 
//////
deff("x = %s_m_UNIT(a,b)","x = b * a"); 

////////
function x = %UNIT_m_UNIT(a, b)
// implements the multiplication between a two unit-vectors
if ~(length(a.values) ==1 | length(b.values) == 1)
  error(10); // inconsistent multiplication...
end ;
if (a.unitSystem == b.unitSystem)
  x = mlist(["UNIT", "unitSystem", "values", "dimension"], a.unitSystem, a.values * b.values , a.dimension + b.dimension);
else
  x = mlist(["UNIT", "unitSystem", "values", "dimension"], a.unitSystem, a.values * b.values * toStdSys(b) / toStdSys(a), a.dimension + b.dimension);
end ;  
endfunction ;

////////
function x = %UNIT_x_UNIT(a, b)
// implements the element-wise multiplication between two unit-vectors
if (length(a.values) ~= length(b.values))
  error(10); // inconsistent multiplication...
end;
if (a.unitSystem == b.unitSystem)
  x = mlist(["UNIT", "unitSystem", "values", "dimension"], a.unitSystem, a.values .* b.values , a.dimension + b.dimension);
else
  x = mlist(["UNIT", "unitSystem", "values", "dimension"], a.unitSystem, a.values .* b.values * toStdSys(b) / toStdSys(a), a.dimension + b.dimension);
end ;
endfunction ;

////////
function x = %UNIT_x_s(a, b)
// implements the element-wise multiplication between a unit-vector and a const-vector
if (length(a.values) ~= length(b))
  error(10); // inconsistent multiplication...
end;
x = mlist(["UNIT", "unitSystem", "values", "dimension"], a.unitSystem, a.values .* b, a.dimension + b.dimension);
endfunction ;

////////
deff("x = %s_x_UNIT(a, b)", "x = b .* a;");		// s .* u
////////

//// POWER /////
////////
function x = %UNIT_p_s(a, b)
// implements a ^b where a is a unit-vector and b is a const scalar
if (length(b)~=1)
  error("The power must be a scalar.");
else
  x = mlist(["UNIT", "unitSystem", "values", "dimension"], a.unitSystem, a.values ^ b, a.dimension * b);
end;
endfunction ;

//// DIVISION /////
deff("x = %UNIT_r_s(a, b)"   , "x = a *  (1 / b);" );	// a / s
deff("x = %s_r_UNIT(a, b)"   , "x = a *  (b^(-1));");	// s / a
deff("x = %UNIT_r_UNIT(a, b)", "x = a * (b^(-1)); ");	// a / b
deff("x = %UNIT_d_UNIT(a, b)", "x = a .* (b^(-1));");	// a ./ b
deff("x = %s_d_UNIT(a, b)"   , "x = a .* (b^(-1));");	// a ./ b

//// ADDITION /////
////////
function x = %UNIT_a_UNIT(a, b)
// implements the addition of two unit-vectors
if ( a.dimension <> b.dimension )
  error("Unmatched dimensions.", 10001);
end;
if ( length(a.values) ~= length(b.values) )
  error(144);
end;
if (a.unitSystem ~= b.unitSystem)
  x = mlist(["UNIT", "unitSystem", "values", "dimension"], SystemList(1), a.values * toStdSys(a) + b.values * toStdSys(b), a.dimension);
  x = x \ UNITSYS(a.unitSystem);
else
  x = mlist(["UNIT", "unitSystem", "values", "dimension"], a.unitSystem, a.values + b.values, a.dimension);
end;
endfunction ;

//////
// Substraction :
deff("x = %UNIT_s_UNIT(a, b)", "x = a + ((-1) * b) ");

//////////////////////
//// COMPARISONS /////
//////////////////////
////////
function x = %UNIT_o_UNIT(a, b)
// implements the boolean test (a == b) where a and b are unit-vectors
if ( a.dimension <> b.dimension )
  error("Unmatched dimensions.", 10001);
end;
if ( length(a.values) ~= length(b.values) )
  error("Vectors must be of the same length.");
end;
if (a.unitSys ~= b.unitSys)
  x = (a.values * toStdSys(a) == b.values * toStdSys(b));
else
  x = (a.values == b.values);
end;
endfunction ;

////////
function x = %UNIT_1_UNIT(a, b)
// implements the boolean test (a < b) where a and b are unit-vectors
if ( a.dimension <> b.dimension )
  error("Unmatched dimensions.", 10001);
end;
if ( length(a.values) ~= length(b.values) )
  error("Vectors must be of the same length.");
end;
if (a.unitSys ~= b.unitSys)
  x = (a.values * toStdSys(a) < b.values * toStdSys(b));
else
  x = (a.values < b.values);
end;
endfunction ;

////////
// Other comparisons :
deff("x = %UNIT_2_UNIT(a, b)", "x = ~( (a < b) | (a == b));"); // a >  b ?
deff("x = %UNIT_3_UNIT(a, b)", "x = ~(a > b);");               // a <= b ?
deff("x = %UNIT_4_UNIT(a, b)", "x = ~(a < b);");               // a >= b ?

////////

//////////////////////////////
//  MAIN CONVERSION FUNCTION :
//////////////////////////////
////////
function x = toStdSys(a);
// Computes the conversion factor allowing to express the unit-vector 'a' (expressed in whatever units) in the reference system of units (the first system in global variable 'SystemList').
if (typeof(a)~="UNIT")
  error("This function is only defined for ""UNIT"" m-lists.");
end;
if ( a.unitSystem == SystemList(1) )
  x = 1 ;
end;
found = %F;
for k = 1 : length(SystemList)
  if (SystemList(k) == a.unitSystem)
    found = %T;
    break ;
  end ;
end;
if ~found
  errorMsg="""" + a.unitSystem + """ system of units not found in scope.";
  error(errorMsg);
end;
command = "Conversions = " + a.unitSystem + ".conversionFactors"; 
execstr(command);
unitSys = UNITSYS(a.unitSystem);
conversionFactors = unitSys.conversionFactors; 
x = prod(conversionFactors .^ a.dimension);
endfunction ;

///////////////////////////////////////////
// INSERTION / EXTRACTION / CONCATENATION : 
///////////////////////////////////////////
////////
function a = %UNIT_i_UNIT(varargin)
// implements insertion "a(i1, 12, ..., in) = b" where 'a' and 'b' are unit-vectors and 'i1',...,'in' are positive const-vectors.
b = varargin($-1)  ;
a = varargin($)    ;
if ( a.dimension <> b.dimension )
  error("Unmatched dimensions.", 10001);
end;
if ((length(a.values(varargin(1:$-2))) ~= length(b.values)) & (length(b.values) ~= 1))
  error(1); // incorrect assignment...
end;
if (a.unitSystem == b.unitSystem)
  a.values(varargin(1:$-2)) = b.values;
else
  a.values(varargin(1:$-2)) = b.values * toStdSys(b) / toStdSys(a);
end;
endfunction;

////////
function x = %UNIT_e(varargin)
// implements extraction "x = a(i1, i2, ..., in)" where 'a' and 'x' are unit-vectors and 'i1',...,'in' are positive const-vectors.
a = varargin($);
x = mlist(["UNIT", "unitSystem", "values", "dimension"], a.unitSystem, a.values(varargin(1:$-1)), a.dimension);
endfunction;

//////////

function x = %UNIT_c_UNIT(a, b)
// implements the horizontal concatenation "x = [a , b]" where 'a', 'b' and 'x' are unit-vectors
if ( a.dimension <> b.dimension )
  error("Unmatched dimensions.", 10001);
end;
if (size(a.values, 1) ~= size(b.values, 1))
  error(5); // inconsistent column/row dimensions...
end;
x = mlist(["UNIT", "unitSystem", "values", "dimension"], a.unitSystem, [a.values, b.values], a.dimension);
endfunction;

////////
function x = %UNIT_f_UNIT(a, b)
// implements the vertical concatenation "x = [a ; b]" where 'a', 'b' and 'x' are unit-vectors
if ( a.dimension <> b.dimension )
  error("Unmatched dimensions.", 10001);
end;
if (size(a.values, 2) ~= size(b.values, 2))
  error(6); // inconsistent row/column dimensions...
end;
x = mlist(["UNIT", "unitSystem", "values", "dimension"], a.unitSystem, [a.values; b.values], a.dimension);
endfunction;




//////////////////////////////
// BUILDING NEW UNIT NUMBERS : 
//////////////////////////////
////////
function x = %s_h_UNIT(a, b)
// implements "a & b" where 'a' and 'x' are unit-vectors. Another operator for scalar x UNIT multiplication, identical to '*'
x = b;
x.values = a * x.values;
endfunction ;

////////
function x = %UNIT_h_UNITSYS(a, b)
// a & b : "Force-casts" the numeric value of unit-vector 'a' to be its value as expressed in the units of the 'b' unit-vector.
x = a; 
x.unitSystem = b.name ;
endfunction ;

////////
function x = %UNIT_l_UNITSYS(a, b)
// implements "a \ b" where 'a' and 'x' are unit-vectors: converts  x in units of the unit-system of unit-vector 'b'.
x = a * toStdSys(a);
x.unitSystem = b.name;
x = x / toStdSys(x);
endfunction ;

////////
function x = UNITSYS(str);
// outputs the unit system in which the unit-vector 'x' is expressed.
execstr("x = " + str);
endfunction;

////////
function x = defQuantity(unitSystem, quantity, unit);
// outputs the unit-vector 'x', declared to be expressed in units of the "unitSystem" system of units, its unit being 'unit' and the numerical value in this system being 'quantity'.
if ( isempty(SystemList))
  error("Define at least one system of units first.");
end;
found = %F;
for k = 1 : length(SystemList)
  if (SystemList(k) == unitSystem)
    found = %T;
    break ;
  end ;
end;
if ~found
  errorMsg="""" + unitSystem + """ system of units not found in scope.";
  error(errorMsg);
end;
U = unit;
S = UNITSYS(SystemList(k));
if (length(unit) < length(S.conversionFactors))
  U($+1:length(S.conversionFactors))=0;
end;
x = mlist(["UNIT", "unitSystem", "values", "dimension"], unitSystem, quantity, fraction(U));
endfunction ;

/////////////////
/////// DISPLAY :
/////////////////
function defVisualUnit(varargin)

// Error checking :
if ((length(varargin) < 2) | (length(varargin) ~= 2 & modulo(length(varargin),2) == 0))
  error(39); // Bad number of arguments...
end;
select typeof(varargin(1))
  case "constant"
    unit = fraction(varargin(1));
  case "FRACTION"
    unit = varargin(1);
  case "UNIT"
    unit = varargin(1).dimension;
  else
    error(44,1); // 1st argument is incorrect...
end;
global SystemList;
if ~exists("SystemList");
  error(" ""SystemList"" not found in scope.\n Define global variable ""SystemList"" before use of this function.");
end;
if isempty("SystemList");
  error("No Unit System is declared. Declare a Unit System before use of this function.");
end;
refSys = UNITSYS(SystemList(1));
if (length(refSys.conversionFactors) ~= length(unit.nums))
  error("the first argument must be a vector of size equal to the number of units declared in the defined Unit Systems.");
end;
global UnitStrings;
for i = 1 : length(UnitStrings)
  if (UnitStrings(i).unit == unit)
    error("Visual unit already defined. Erase it by hand first by modifying ""UnitStrings"".");
  end;
end;

// Execution:
if (length(varargin) == 2)
  UnitStrings($ + 1) = mlist(["UNITSTRING", "unit", "unitSystems", "strings"], unit, [], []);
  for i = 1 : length(SystemList)
    UnitStrings($).strings(i) = varargin(2);
    UnitStrings($).unitSystems(i) = SystemList(i);
  end;
else
  UnitStrings($ + 1) = mlist(["UNITSTRING", "unit", "unitSystems", "strings"], unit, [], []);
  for i = 1 : floor((length(varargin) - 1)/2)
    UnitStrings($).strings(i) = varargin(1 + 2 * i);
    UnitStrings($).unitSystems(i) = varargin(2 * i).name;
  end;
end;
endfunction;

////////
function str = prettyPrint(a);

str = "";
for i = 1 : length(UnitStrings)
  if (UnitStrings(i).unit == a.dimension)
    for j = 1:length(length(UnitStrings(i).strings))
      if (a.unitSystem == UnitStrings(i).unitSystems(j))
         str = UnitStrings(i).strings(j);
         return;
      end;
    end;
  end;
end;
endfunction;


////////
function %UNIT_p(a)
// displaying rule for a unit-vector:
separatorChar= code2str(283);   // the character "" will separate the units.
unitSys     = UNITSYS(a.unitSystem);
unitStrings  = unitSys.unitNames;
printedUnits = [""];
separators   = [""];
lastUnit = 0;
for k = 1 : length(unitSys.conversionFactors);
  if (a.dimension.nums(k) == 0) then
    printedExps(k)  = ""; 
    printedUnits(k) = "";
    continue;
  elseif (a.dimension.nums(k) < 0)
    lastUnit = k;
    printedUnits(k) = msprintf("%s", unitStrings(k));
    unitLength = length(printedUnits(k));
    select a.dimension.denums(k)
    case 1 then
      printedExps(k) = msprintf("%i", a.dimension.nums(k));
    else
      printedExps(k) = msprintf("%i/%i", a.dimension.nums(k), a.dimension.denums(k));
    end;
  else
    lastUnit = k;
    printedUnits(k) = msprintf("%s", unitStrings(k));
    unitLength = length(printedUnits(k));
    select a.dimension.denums(k)
    case 1 then
      if (a.dimension.nums(k) ~= 1) 
        printedExps(k) = msprintf("%i", a.dimension.nums(k));
      else
        printedExps(k) = "";
      end;
    else
      printedExps(k)   = msprintf("%i/%i", a.dimension.nums(k), a.dimension.denums(k));
    end;
  end;
  // More formatting:
  expLength = length(printedExps(k));


  for l = 1:unitLength do printedExps(k)  = " " + printedExps(k) ; end;
  for l = 1:expLength  do printedUnits(k) = printedUnits(k) + " "; end;
  printedUnits(k) = printedUnits(k) + separatorChar;
  printedExps(k)  = printedExps(k)  + " ";
end;

if (exists("PRETTYPRINTEDUNIT") & PRETTYPRINTEDUNIT == %T)
  prettyPrintedUnit = prettyPrint(a) + " ";
else
  prettyPrintedUnit = "";
end;

prettyPrintedSpacer = "";
for l = 1 : length(prettyPrintedUnit)
  prettyPrintedSpacer = prettyPrintedSpacer + " ";
end;

if (lastUnit > 0) 
  // erase the last separator at the end of the "dimension" string
  printedUnits(lastUnit) = strsubst(printedUnits(lastUnit), separatorChar,"");
end;
if (length(a.values) == 1)
  // if 'a' is a scalar, print the value and the dimension on the same line in the console.
  printf(   "\t\t%s \n%10.6g\t%s   (%s)\n", prettyPrintedSpacer + strcat(printedExps), a.values, prettyPrintedUnit + strcat(printedUnits), a.unitSystem);
  return;
else
  // for vectors, print all the values as regular CONST vectors first, then new line and print the dimension.
  disp(a.values);
  printf("\n\t%s\nUnit:\t%s   (%s)\n", prettyPrintedSpacer + strcat(printedExps), prettyPrintedUnit + strcat(printedUnits), a.unitSystem);
  return;
end;

endfunction;

//////////////////////
// UTILITY FUNCTIONS :
//////////////////////
////////
function x = len(a)
// length of a unit-vector
x = length(a.values);
endfunction;

////////
function x = siz(varargin)
//size of a unit-vector (which can actually have, as const-vectors, any number of dimensions).
a = varargin(1);
x = size(a.values, varargin(2:$));
endfunction;

////////
function x = numeric(a);
// Returns the numerical values contained in the unit-vector 'a' as an array of const. (mandatory if you want to use your unit-vectors on any Scilab function, such as log, sin, plot, ...)
if ( typeof(a)~="UNIT" )
  error("This function is only defined for ""UNIT"" m-lists.");
end;
x = double(a.values);
endfunction;

///////////


// The following is a simple redefinition of the fraction type. 
// Fractions are used to represent the unit of the physical quantities described by the so-called "unit-vectors".
// The operators allow to multiply, exponentiate, add, ... physical quantities and also generate the appropriate errors when
// one tries to do non-physical operations (such as add two quantities of different physical meaning).
// For instance : "1 kelvin + 1 meter" would throw an 10001 exception (Inconsistent expression...).



/////////////////////////////////////////////
// FRACTION REDEF. FOR EXPONENT CALCULATION :
/////////////////////////////////////////////

function x = fraction(a)
// approximates (when possible) a real number 'a' as a rational number with a denumerator inferior to 100. 
// This function is specially designed to make the definition of physical quantites easier, 
// where the exponents of units are "simple" fractions (such as m^2, m^(1/3)s^(-1), ...)
Nums   = zeros(1, length(a));
Denums = zeros(1, length(a));
for k = 1 : length(a)
  if (a(k) == 0)
    Nums(k)   = int16(0);
    Denums(k) = int16(1);
    continue; 
  end;
  Denum = 1;
  rationalError = 0;
  while (Denum < 100)
    if (abs(double(a(k) * Denum) - round(double(a(k) * Denum))) < 10 * %eps) 
      break;
    else
      rationalError(Denum) = abs(double(a(k) * Denum) - round(double(a(k) * Denum))) ;
    end;
    Denum = Denum + 1;
  end;
  if (Denum == 100) 
    [m, Denum] = min(rationalError);
    warnMsg = msprintf("Couldn''t find a suitable fraction with a denumerator inferior to 100 for %f (element #%i). Best rational approximation is %i/%i (error : %e).", a(k), k, int16(Denum * a(k)), Denum, abs(m));
    warning(warnMsg);
    g = gcd(int16([Denum, round(Denum * a(k))]));
  else
    g = 1;
  end;
  Nums(k)   = int16(round(Denum * a(k)) / g);
  Denums(k) = int16(round(Denum));
  if (Denums(k) < 0)
    Denums(k) = -Denums(k);
    Nums(k)   = -Nums(k);
  end;
end;
x = mlist(["FRACTION", "nums", "denums"], Nums, Denums);
endfunction ;

////////
function x = %FRACTION_a_s(a, b)
// implements the addition of a fraction-vector(/scalar) 'a' and a const-scalar(/vector) 'b'
if ((length(a.nums) ~= 1) & (length(b)~=1) & (length(a.nums) ~= length(b)))
  error(8); // inconsistent addition...
end;
b = fraction(b);
if (length(b.nums) > length(a.nums))
  // operation of type fraction_vector + const_scalar:
  for k = 1 : length(b.nums)
    g(k) = gcd(int16([a.nums * b.denums(k) + a.denums * b.nums(k), a.denums * b.denums(k)]));
  end;
  x = mlist(["FRACTION", "nums", "denums"], int16((a.nums + a.denums *  b.nums) ./ abs(g)), int16(b.denums ./ abs(g)));
  return;
else
  for k = 1 : length(a.nums)
    g(k) = gcd(int16([a.nums(k) * b.denums + a.denums(k) * b.nums, a.denums(k) * b.denums]));
  end;
  if (length(a.nums) ~= length(b.nums))
    // operation of type fraction_vector + const_scalar:
    x = mlist(["FRACTION", "nums", "denums"], int16((a.nums + a.denums *  b.nums) ./ abs(g')), int16(a.denums ./ abs(g')));
    return;
  else
    // operation of type fraction_vector + fraction_vector:
    x = mlist(["FRACTION", "nums", "denums"], int16((a.nums + a.denums .* b.nums) ./ abs(g')), int16(a.denums ./ abs(g')));
    return;
  end;
end;
endfunction ;

////////

deff("x = %s_a_FRACTION(a, b)", "x = b + a;");
deff("x = %FRACTION_s_s(a, b)", "x = a + (-b);");
deff("x = %s_s_FRACTION(a, b)", "x = b + (-a);");

////////
function %FRACTION_p(a)
// Displaying rule for a fraction:
for k = 1 : length(a.nums) do 
  printf("  (%i)\n", k);
  if (a.denums(k) ~= 1)
    printf("    %i / %i\n", a.nums(k), a.denums(k)); 
  else
    printf("    %i\n", a.nums(k));
  end;
end; 
endfunction;

////////
function x = %FRACTION_m_s(a, b)
// implements the multiplication between a fraction-vector(/scalar) 'a' and a const-scalar(/vector) 'b'
if ((length(a.nums) ~= 1) & (length(b)~=1) )
  error(10); // inconsistent multiplication...
end; 
b = fraction(b);  
Num   = zeros(1, length(a.nums));
Denum = zeros(1, length(a.nums));
if (length(a.nums) >= length(b.nums))
  for k = 1 : length(a.nums)
    Num(k)   = a.nums(k) * b.nums;
    Denum(k) = a.denums(k) * b.denums;
    g = gcd(int16([Num(k), Denum(k)]));
    Num(k)   = int16(Num(k)   / g);
    Denum(k) = int16(Denum(k) / g);
  end;
else
  for k = 1 : length(b.nums)
    Num(k)   = a.nums   * b.nums(k);
    Denum(k) = a.denums * b.denums(k);
    g = gcd(int16([Num(k), Denum(k)]));
    Num(k)   = int16(Num(k)   / g);
    Denum(k) = int16(Denum(k) / g);
  end;
end;
sgn = sign(Denum);
x = mlist(["FRACTION", "nums", "denums"], Num .* sgn, Denum .* sgn);
endfunction ;

////////

deff("x = %s_m_FRACTION(a, b)", "x = b * a");
deff("x = %s_r_FRACTION(a, b)", "sgn = sign(b.nums); x = a * mlist([""FRACTION"", ""nums"", ""denums""], b.denums .* sgn, b.nums .* sgn)");
deff("x = %FRACTION_r_s(a, b)", "x = a * (1/b)");

////////
function x = %FRACTION_m_FRACTION(a, b)
// implements the multiplication between a fraction-vector(/scalar) 'a' and a fraction-scalar(/vector) 'b'
if ((length(a.nums) ~= 1) & (length(b.nums)~=1) )
  error(10); // inconsistent multiplication...
end;
if (length(b.nums) ~= 1)
  x = b * a;
  return;
end;
// from here on, b is a singleton;
Nums   = zeros(1, length(a.nums));
Denums = zeros(1, length(a.nums));
for k = 1 : length(a.nums) 
  if (a.denums(k) * b.denums == 0)
    error(27); // Division by zero...
  end;
  if ((a.nums(k) == 0) | (b.nums == 0))
    Nums(k)   = int16(0);
    Denums(k) = int16(1);
  end;   
  Num   = a.nums(k)   * b.nums;
  Denum = a.denums(k) * b.Denums;
  g = gcd(int16([Num, Denum]));
  Nums(k)   = int16(Num / g);
  Denums(k) = int16(Denum / g);
end;
x = mlist(["FRACTION", "nums", "denums"], Nums, Denums);
endfunction ;

////////////

deff("x = %FRACTION_r_FRACTION(a, b)","sgn = sign(b.nums); x = a * mlist([""FRACTION"", ""nums"", ""denums""], b.denums .* sgn, b.nums .* sgn)");

////////////

function x = %FRACTION_x_FRACTION(a, b)
// implements the piecewise multiplication "a .* b" between two fraction-vectors 'a' and 'b'
if (length(a.nums) ~= length(b.nums))
  error(10); // inconsistent multiplication...
end;
if or((a.denums .* b.denums) == 0)
  error(27); // division by zero...
end;
g = zeros(1, length(a.nums));
for k = 1 : length(a.nums) 
  g(k) = gcd(int16([a.nums(k) * b.nums(k), a.denums(k) * b.denums(k)]));
end;
x = mlist(["FRACTION", "nums", "denums"], int16(a.nums .* b.nums ./ abs(g)), int16(a.denums .* b.denums ./ abs(g)));
endfunction ;

////////////

deff("x = %FRACTION_x_s(a, b)", "x = a .* fraction(b)");

////////////

function x = %FRACTION_d_FRACTION(a, b)
// implements the piecewise division "a ./ b" between two fraction-vectors 'a' and 'b'
sgn = sign(b.nums); 
if (or(sgn == 0))
  error(27); // division by zero...
end;
x = a .* mlist(["FRACTION", "nums", "denums"], b.denums .* sgn, b.nums .* sgn);
endfunction;

////////
function x = %FRACTION_a_FRACTION(a, b)
// implements the piecewise addition "a + b" between two fraction-vectors 'a' and 'b'
if (length(a.nums) ~= length(b.nums))
  error("Fractions must be vectors of the same length.");
end;
Num   = a.nums .* b.denums + a.denums .* b.nums;
Denum = a.denums .* b.denums;
for k = 1:length(a.nums)
  g = abs(gcd(int16([Num(k), Denum(k)])));
  Num(k)   = int16(Num(k) / g);
  Denum(k) = int16(Denum(k) /g);
end;
x = mlist(["FRACTION", "nums", "denums"], Num, Denum);
endfunction ;

////////

deff("x = %FRACTION_s_FRACTION(a, b)", "x = a + (-1) * b");

////////
function x = %s_p_FRACTION(a, b)
// Implements a ^ b, where a is a const-vector and b is a fraction-vector.
// x is of const type (this function is primarily used for conversion between Units Systems).
x = a ^ (double(b.nums) / double(b.denums) );
endfunction;

////////
function x = value(a)
// Returns the const numerical value of fraction-vector 'a'
if (typeof(a)~="FRACTION")
  error("Wrong argument: function is only defined for ""FRACTION""-type m-lists.");
end;
x = double(a.nums) ./ double(a.denums);
endfunction ;

////////
function x = %s_j_FRACTION(a, b)
// Implements the piecewise exponentiation a .^ b where 'a' and 'b'  both are fraction-vectors. 
// (useful for conversions between Systems of Units).
x = a .^ value(b);
endfunction;

////////
function x = %FRACTION_n_FRACTION(a, b)
// Implements the boolean test (a <> b) where 'a' and 'b' are both fraction-vectors.
// x = %t if for every k, a(k) == b(k). x = %f else.
x = (sum(value(a - b) > 0) > 0);
endfunction;

////////////
funcprot(1);